@@ -10,11 +10,16 @@ from pathlib import Path
 
 import ghidra_bridge
 
+from mercury_engine_data_structures import dread_data
+
 hash_str = "HashString"
 register_field = "RegisterField"
+add_enum_value = "(?:FUN_71000148b8|AddEnumValue)"
 prefixes_to_remove = [
     "(ObjectField *)",
     "&",
+    "(CClass *)",
+    "Reflection::"
 ]
 
 _aliases = {
@@ -81,13 +86,25 @@ def clean_crc_var(crc_var: str) -> str:
     return crc_var
 
 
+def fix_alternative_ghidra_name(name: str) -> str:
+    if name.endswith("Ptr"):
+        name = name[:-len("Ptr")] + "*"
+    name = name.replace("_const", " const")
+    name = name.replace(",_", ", ")
+    name = name.replace("Ptr>", "*>")
+    return name
+
+
 def get_field_registrations(bridge: ghidra_bridge.GhidraBridge, ifc, monitor, fields_function):
+    if fields_function is None:
+        return {}
+
     res = bridge.remote_eval("""
         ifc.decompileFunction(fields_function, 180, monitor)
     """, timeout_override=200, fields_function=fields_function, ifc=ifc, monitor=monitor)
 
     decompiled_code = str(res.getCCodeMarkup())
-    hash_call_re = re.compile(hash_str + r'\(([^,]+?),"?([^,]+?)"?,1\);')
+    hash_call_re = re.compile(hash_str + r'\(([^,]+?),"?([^,]+?)"?,(?:1|true)\);')
     register_call_re = re.compile(register_field + r'\([^,]+?,([^,]+?),(.+?),([^,]+?),([^,]+?),([^,]+?)\);')
 
     crc_mapping = collections.defaultdict(list)
@@ -106,46 +123,111 @@ def get_field_registrations(bridge: ghidra_bridge.GhidraBridge, ifc, monitor, fi
             if offset < m.start():
                 break
 
+        if crc_string is None:
+            raise ValueError(f"Could not find the correct string for {crc_var}")
+
         if "&" in type_var:
             if "::_" in type_var:
-                type_name = type_var[1:type_var.find("::_")]
+                type_name = type_var[type_var.find("&") + 1:type_var.find("::_")]
             else:
                 type_name = type_var
         else:
             i = decompiled_code.rfind(type_var, offset, m.start())
             end = decompiled_code.find(';', i)
             type_name = decompiled_code[i + len(type_var) + len(" = "):end]
-            if type_name.endswith("::init()"):
-                type_name = type_name[:-len("::init()")]
+            for prefix in prefixes_to_remove:
+                if type_name.startswith(prefix):
+                    type_name = type_name[len(prefix):].strip()
+            if type_name.endswith("()"):
+                type_name = type_name[:-len("()")]
+            if type_name.endswith("::init"):
+                type_name = type_name[:-len("::init")]
 
         fields[crc_string] = _aliases.get(type_name, type_name)
 
     return fields
 
 
-def get_function_list() -> dict[str, tuple[int, int]]:
+def get_value_registrations(bridge: ghidra_bridge.GhidraBridge, ifc, monitor, values_function):
+    if values_function is None:
+        return None
+
+    res = bridge.remote_eval("""
+        ifc.decompileFunction(values_function, 180, monitor)
+    """, timeout_override=200, values_function=values_function, ifc=ifc, monitor=monitor)
+
+    decompiled_code = str(res.getCCodeMarkup())
+    hash_call_re = re.compile(hash_str + r'\(([^,]+?),"?([^,]+?)"?,(?:1|true)\);')
+    enum_call_re = re.compile(add_enum_value + r'\([^,]+?,([^,]+?),(.+?)\);')
+
+    crc_mapping = collections.defaultdict(list)
+    values = {}
+
+    for m in hash_call_re.finditer(decompiled_code):
+        crc_var, crc_string = m.group(1, 2)
+        crc_mapping[clean_crc_var(crc_var)].append((m.start(), crc_string))
+
+    for m in enum_call_re.finditer(decompiled_code):
+        crc_var, value_var = m.group(1, 2)
+
+        offset = None
+        crc_string = None
+        for offset, crc_string in reversed(crc_mapping[clean_crc_var(crc_var)]):
+            if offset < m.start():
+                break
+
+        if crc_string is None:
+            raise ValueError(f"Could not find the correct string for {crc_var}")
+
+        values[crc_string] = int(value_var, 0)
+
+    return values
+
+
+def get_function_list() -> dict[str, tuple[int, int, int]]:
     with ghidra_bridge.GhidraBridge() as init_bridge:
-        result = init_bridge.remote_eval("""
+        result_fields = init_bridge.remote_eval("""
         [
-            (f.getName(True), f.getID()) for f in currentProgram.getSymbolTable().getDefinedSymbols()
-            if f.getName() == "fields" or f.getName() == "init"
+            (f.getName(True), f.getID()) for f in currentProgram.getSymbolTable().getSymbols("fields")
         ]
         """)
+        result_init = init_bridge.remote_eval("""
+        [
+            (f.getName(True), f.getID()) for f in currentProgram.getSymbolTable().getSymbols("init")
+        ]
+        """)
+        result_values = init_bridge.remote_eval("""
+        [
+            (f.getName(True), f.getID()) for f in currentProgram.getSymbolTable().getSymbols("values")
+        ]
+        """)
+
+        print("Found {} init, {} fields and {} values".format(
+            len(result_init),
+            len(result_fields),
+            len(result_values),
+        ))
+
         init_funcs = {}
         fields_funcs = {}
-        for name, func_id in result:
-            if not name.startswith("Reflection::"):
+        values_funcs = {}
+        for name, func_id in result_fields + result_init + result_values:
+            if name.startswith("Reflection::"):
+                name = name[len("Reflection::"):]
+
+            if name.startswith("base::reflection::CollectionTypeMapper"):
                 continue
-            name = name[len("Reflection::"):]
 
             if name.endswith("::init"):
                 init_funcs[name[:-len("::init")]] = func_id
             elif name.endswith("::fields"):
                 fields_funcs[name[:-len("::fields")]] = func_id
+            elif name.endswith("::values"):
+                values_funcs[name[:-len("::values")]] = func_id
 
         return {
-            name: (init_funcs.get(name), fields_funcs[name])
-            for name in fields_funcs
+            name: (init_funcs.get(name), fields_funcs.get(name), values_funcs.get(name))
+            for name in fields_funcs.keys() | init_funcs.keys() | values_funcs.keys()
         }
 
 
@@ -163,12 +245,16 @@ def initialize_worker():
 
     monitor = ConsoleTaskMonitor()
     ifc = DecompInterface()
-    ifc.setOptions(DecompileOptions())
+    options = DecompileOptions()
+    # make sure you have namespaces set to Always, and casting disabled
+    options.grabFromProgram(flat_api.currentProgram)
+    ifc.setOptions(options)
     ifc.openProgram(flat_api.currentProgram)
 
 
-def decompile_type(type_name: str, init_id: typing.Optional[int], fields_id: int,
-                   ) -> tuple[str, typing.Optional[str], dict[str, str]]:
+def decompile_type(type_name: str, init_id: typing.Optional[int], fields_id: typing.Optional[int],
+                   values_id: typing.Optional[int]
+                   ) -> tuple[str, typing.Optional[str], dict[str, str], dict[str, int]]:
     if bridge is None:
         raise ValueError("Bridge not initialized")
 
@@ -188,23 +274,34 @@ def find_parent(f):
             )
         )""", func_id=init_id)
 
-    func = bridge.remote_eval("""
-        currentProgram.getFunctionManager().getFunctionAt(
-            currentProgram.getSymbolTable().getSymbol(func_id).getAddress()
-        )
-    """, func_id=fields_id)
+    func = None
+    if fields_id is not None:
+        func = bridge.remote_eval("""
+            currentProgram.getFunctionManager().getFunctionAt(
+                currentProgram.getSymbolTable().getSymbol(func_id).getAddress()
+            )
+        """, func_id=fields_id)
 
     fields = get_field_registrations(bridge, ifc, monitor, func)
 
+    func = None
+    if values_id is not None:
+        func = bridge.remote_eval("""
+            currentProgram.getFunctionManager().getFunctionAt(
+                currentProgram.getSymbolTable().getSymbol(func_id).getAddress()
+            )
+        """, func_id=values_id)
+    values = get_value_registrations(bridge, ifc, monitor, func)
+
     if parent_init is not None:
         if parent_init.startswith("Reflection::"):
             parent_init = parent_init[len("Reflection::"):]
         parent_init = parent_init[:-len("::init")]
 
-    return type_name, parent_init, fields
+    return type_name, parent_init, fields, values
 
 
-def decompile_in_background(all_fields_functions: dict[str, tuple[int, int]]):
+def decompile_in_background(all_fields_functions: dict[str, tuple[int, int, int]]):
     process_count = max(multiprocessing.cpu_count() - 2, 2)
 
     finished_count = 0
@@ -222,8 +319,8 @@ def decompile_in_background(all_fields_functions: dict[str, tuple[int, int]]):
     result = {}
 
     def callback(r):
-        type_name, parent, fields = r
-        result[type_name] = {"parent": parent, "fields": fields}
+        type_name, parent, fields, values = r
+        result[type_name] = {"parent": parent, "fields": fields, "values": values}
         report_update(f"Parsed {type_name}")
 
     if total_count > process_count:
@@ -260,7 +357,39 @@ def decompile_in_background(all_fields_functions: dict[str, tuple[int, int]]):
     return result
 
 
-def main(only_missing: bool = True):
+def is_invalid_data(name: str, data: dict[str, typing.Any]):
+    # if "null" in data["fields"]:
+    #     return True
+
+    # if data["values"] is not None and "null" in data["values"]:
+    #     return True
+
+    return False
+
+
+def is_container_or_ptr(name: str):
+    prefixes = [
+        "base::global::CRntSmallDictionary",
+        "base::global::CRntDictionary",
+        "base::global::CRntVector",
+        "base::global::CWeakPtr",
+        "base::global::CSmartPtr",
+        "std::unique_ptr",
+    ]
+    suffixes = [
+        "Ptr",
+        "::value_type",
+        "::TKeyElementIterator",
+        "::TKeyElementConstIterator",
+        "::TElementIterator",
+        "::TElementConstIterator",
+    ]
+
+    return any(name.endswith(suffix) for suffix in suffixes) or any(name.startswith(prefix) for prefix in prefixes)
+
+
+def main(only_missing: bool = True, ignore_without_hash: bool = True,
+         ignore_existing_invalid_fields: bool = True, ignore_container_or_ptr: bool = True):
     print("Getting function list")
     all_fields_functions = get_function_list()
     print(f"Got {len(all_fields_functions)} functions!")
@@ -273,10 +402,39 @@ def main(only_missing: bool = True):
     except (FileNotFoundError, json.JSONDecodeError):
         final_results = {}
 
+    final_results = typing.cast(dict[str, typing.Any], final_results)
+
+    if ignore_container_or_ptr:
+        for key in list(all_fields_functions.keys()):
+            if is_container_or_ptr(key):
+                all_fields_functions.pop(key)
+
+    if ignore_existing_invalid_fields:
+        for key in list(final_results.keys()):
+            invalid = is_invalid_data(key, final_results[key])
+            container = ignore_container_or_ptr and is_container_or_ptr(key)
+            unknown_hash = False
+            # unknown_hash = ignore_without_hash and key not in dread_data.all_name_to_property_id()
+
+            if invalid or container or unknown_hash:
+                print(f"Removing existing type: {key}")
+                final_results.pop(key)
+
     if only_missing:
         for key in final_results.keys():
             all_fields_functions.pop(key, None)
 
+    if ignore_without_hash:
+        # for key in list(final_results.keys()):
+        #     if key not in dread_data.all_name_to_property_id():
+        #         print(f"Removing {key}: no known hash")
+        #         final_results.pop(key)
+
+        for key in list(all_fields_functions.keys()):
+            if key not in dread_data.all_name_to_property_id():
+                # print(f"Skipping {key}: no known hash - {all_fields_functions[key]}")
+                all_fields_functions.pop(key)
+
     process_results = decompile_in_background(all_fields_functions)
     for key in sorted(process_results.keys()):
         final_results[key] = process_results[key]
@@ -288,14 +446,18 @@ def main(only_missing: bool = True):
                 value = _aliases[value]
 
             if not value.startswith("&"):
-                if value.endswith("Ptr"):
-                    value = value[:-len("Ptr")] + "*"
-                value = value.replace("_const", " const")
-                value = value.replace(",_", ", ")
-                value = value.replace("Ptr>", "*>")
+                value = fix_alternative_ghidra_name(value)
 
             data["fields"][field] = value
 
+    for key in list(final_results.keys()):
+        # Something causes a type to inherit from a pointer to itself, that's wrong
+        if final_results[key]["parent"] in (f"{key}Ptr", key):
+            print(f'Removing parent for {key}: {final_results[key]["parent"]}')
+            final_results[key]["parent"] = None
+
+    _merge_split_types(final_results)
+
     with path.open("w") as f:
         json.dump({
             key: final_results[key]
@@ -303,11 +465,50 @@ def main(only_missing: bool = True):
         }, f, indent=4)
 
 
+def _merge_split_types(final_results: dict[str, typing.Any]):
+    hashes = dread_data.all_name_to_property_id()
+    wrong_to_correct = {}
+
+    for key in list(final_results.keys()):
+        if len(key.split("::")) == 1:
+            continue
+
+        last_part = key.split("::")[-1]
+        if last_part.startswith("E") or last_part in {"SKey", "CParams", "SState", "CDefinition", "SSubState",
+                                                      "CActorDef"}:
+            continue
+
+        similar = [other for other in final_results.keys() if other.endswith(last_part) and other != key]
+        if similar:
+            if len(similar) != 1:
+                raise ValueError(f"OH NO too many similar {similar}")
+
+            both_hash = key in hashes and similar[0] in hashes
+            if both_hash:
+                print("{} (Hash: {}) is similar to {} (Hash: {})".format(
+                    key, key in dread_data.all_name_to_property_id(),
+                    similar[0], similar[0] in dread_data.all_name_to_property_id(),
+                ))
+            elif key in hashes:
+                wrong_to_correct[similar[0]] = key
+            elif similar[0] in hashes:
+                wrong_to_correct[key] = similar[0]
+            else:
+                print(f"BOTH ARE WRONG! {key} - {similar[0]}")
+
+    for wrong, correct in wrong_to_correct.items():
+        final_results[correct]["fields"] = final_results.pop(wrong)["fields"]
+
+    for key in list(final_results.keys()):
+        if final_results[key]["parent"] in wrong_to_correct:
+            final_results[key]["parent"] = wrong_to_correct[final_results[key]["parent"]]
+
+
 def simple_decompile():
     all_fields_functions = get_function_list()
     initialize_worker()
 
-    func_name = "CComponent"
+    func_name = "EBreakableTileType"
     print(decompile_type(func_name, *all_fields_functions[func_name]))
 
     # print(decompile_function(*all_fields_functions[4]))
